"""
This module contains a subset of //tobikodata/tcloud/auth.py that only is
concerened with the ability to load and refresh a tcloud JWT token.  It also
providers a PublicHttpClient child class that uses tcloud JWT tokens to
authenticate.  There is code duplication between this and tcloud due to the
isolation that tcloud needs.
"""

import os
import stat
import time
import typing as t
from pathlib import Path

from authlib.integrations.requests_client import OAuth2Session
from httpx import Auth, HTTPStatusError, Request, Response
from rich.console import Console
from rich.theme import Theme
from ruamel.yaml import YAML

from tobikodata.http_client.public import PublicHttpClient

# Yaml
yaml = YAML()

# This is duplicated from tcloud in order to avoid pulling in tcloud deps into
# http client
TCLOUD_PATH = Path(os.environ.get("TCLOUD_HOME", Path.home() / ".tcloud"))
"""The location of the tcloud config folder"""

CLIENT_ID = "f695a000-bc5b-43c2-bcb7-8e0179ddff0c"
"""The oauth client ID to use"""

TOKEN_URL = "https://cloud.tobikodata.com/auth/token"
"""The oauth token endpoint to use"""

THEME = Theme(
    {
        "info": "bright_white",
        "dim": "white",
        "error": "red",
        "success": "green",
        "url": "bright_blue",
        "key": "bright_yellow",
        "provider": "bright_magenta",
        "tobiko": "#FF5700",
    }
)
"""The Rich console theme to use in the CLI"""


class SSOAuth:
    """
    This class handles the OAuth flows and CLI process for refreshing an ID
    Token from tcloud.  Authentication initially must be done with tcloud.
    """

    @staticmethod
    def _auth_yaml_path() -> Path:
        if not TCLOUD_PATH.exists():
            TCLOUD_PATH.mkdir(parents=True, exist_ok=True)
        return TCLOUD_PATH / "auth.yaml"

    @staticmethod
    def _delete_auth_yaml() -> None:
        """
        Removes the auth.yaml file if it exists.
        """
        auth_path = SSOAuth._auth_yaml_path()

        if auth_path.exists():
            os.remove(auth_path)

    @staticmethod
    def _load_auth_yaml() -> t.Optional[t.Dict]:
        """
        Loads the full auth.yaml file that might exist in the CLI config folder.
        """
        auth_path = SSOAuth._auth_yaml_path()

        if auth_path.exists():
            with auth_path.open("r") as fd:
                return yaml.load(fd.read())

        return None

    @staticmethod
    def _save_auth_yaml(data: t.Dict) -> None:
        """
        Saves the given dictionary to auth.yaml

        Args:
            data: The dictionary to save
        """
        auth_path = SSOAuth._auth_yaml_path()

        with auth_path.open("w") as fd:
            yaml.dump(data, fd)
        os.chmod(auth_path, stat.S_IWUSR)

    def __init__(self) -> None:
        self.console = Console(theme=THEME)
        self.session = OAuth2Session(CLIENT_ID)
        self.tokenInfo = SSOAuth._load_auth_yaml()

    def id_token(self) -> t.Optional[str]:
        """
        Returns the id_token needed for SSO.  Will return the one saved on disk,
        unless it's expired.  If the token on disk is expired, it will try to
        refresh it.
        """

        if self.tokenInfo:
            # If we are within 5 minutes of expire time, run refresh
            if self.tokenInfo.get("expires_at", 0.0) > (time.time() + 300):
                # We have a current token on disk, return it
                return self.tokenInfo["id_token"]

            else:
                # Our token is expired, refresh it if possible
                try:
                    refreshed_token = self.refresh_token()

                    if refreshed_token:
                        return refreshed_token

                    # We failed to refresh, logout
                    SSOAuth._delete_auth_yaml()

                except Exception:
                    # We failed to refresh, logout
                    SSOAuth._delete_auth_yaml()

        return None

    def refresh_token(self) -> t.Optional[str]:
        if not self.tokenInfo:
            self.console.print("Not currently authenticated", style="error")
            return None

        current_refresh_token = self.tokenInfo["refresh_token"]

        if not current_refresh_token:
            self.console.print("Refresh token not available", style="error")
            return None

        self.console.print(
            "[info]Refreshing your authentication token[/info] :arrows_counterclockwise:"
        )
        self.session.refresh_token(
            TOKEN_URL, refresh_token=current_refresh_token, scope=self.tokenInfo["scope"]
        )

        return self._create_token_info(self.session.token)["id_token"]

    def _create_token_info(self, token: t.Dict) -> t.Dict:
        self.tokenInfo = {
            "scope": token["scope"],
            "token_type": token["token_type"],
            "expires_at": token["expires_at"],
            "access_token": token["access_token"],
            "refresh_token": token["refresh_token"],
            "id_token": token["id_token"],
        }

        SSOAuth._save_auth_yaml(self.tokenInfo)

        return self.tokenInfo


class SSORequestAuth(Auth):
    def __init__(self, id_token: str) -> None:
        self.id_token = id_token

    def auth_flow(self, request: Request) -> t.Generator[Request, Response, None]:
        request.headers["Authorization"] = f"Bearer {self.id_token}"
        yield request


UNAUTHENTICATED_MESSAGE = 'You are not authenticated. Please either specify a token in your project config, or run "tcloud auth login" if using Tobiko Cloud single sign on.'


class AuthHttpClient(PublicHttpClient):
    """
    This client retries on authentication failures using the TCloud token if available
    """

    def __init__(self, sso: t.Optional[SSOAuth] = None, *args: t.Any, **kwargs: t.Any):
        self.sso = sso or SSOAuth()
        super().__init__(*args, **kwargs)

    def _call(self, *args: t.Any, **kwargs: t.Any) -> Response:
        old_auth = self._client.auth
        has_auth = old_auth or ("headers" in kwargs and "Authorization" in kwargs["headers"])

        if not has_auth:
            # Since we don't have a token configured, let's see if we already
            # have a token, if so use it
            id_token = self.sso.id_token()
            if id_token:
                self._client.auth = SSORequestAuth(id_token)

        try:
            response = super()._call(*args, **kwargs)
            if response.status_code == 401:
                response.raise_for_status()
            return response
        except HTTPStatusError as e:
            if e.response.status_code == 401:
                raise HTTPStatusError(
                    UNAUTHENTICATED_MESSAGE, request=e.request, response=e.response
                )
            raise e
        except self.error_class as e:
            if e.status_code == 401:
                raise self.error_class(UNAUTHENTICATED_MESSAGE, status_code=401)
            raise e
        finally:
            self._client.auth = old_auth



from enum import Enum


class V1Status(str, Enum):
    SCHEDULED = "scheduled"
    IN_PROGRESS = "in_progress"
    SKIPPED = "skipped"
    SUCCESS = "success"
    FAILED = "failed"

    @property
    def complete(self) -> bool:
        return self.value in (V1Status.SUCCESS, V1Status.FAILED, V1Status.SKIPPED)



import typing as t
from functools import cached_property

from tobikodata.helpers import urljoin
from tobikodata.http_client import HttpClient, HttpClientError
from tobikodata.http_client.api_models.v1.dags import V1Dag
from tobikodata.http_client.api_models.v1.evaluations import V1EvaluationBase, V1RunEvaluation
from tobikodata.http_client.api_models.v1.runs import V1Run


class _Base:
    def __init__(self, client: HttpClient, endpoint: str):
        self.client = client
        self.endpoint = endpoint


class V1DagApi(_Base):
    def get_dag_for_environment(self, environment: str) -> t.Optional[V1Dag]:
        """
        Return a DAG object representing the snapshot dependency graph for the specified environment name

        Args:
            environment: The environment name

        Returns:
            The dependency graph, or None if the environment doesnt exist
        """
        try:
            return self.client.get_deserialized(model=V1Dag, url_parts=[self.endpoint, environment])
        except HttpClientError as e:
            if e.status_code == 404:
                return None
            raise


class V1RunApi(_Base):
    def get_last_run_for_environment(self, environment: str) -> t.Optional[V1Run]:
        """
        Fetch information about the last run that occurred in the specified environment

        Args:
            environment: The environment name

        Returns:
            The last run, or None if the environment doesnt exist
        """
        return next(iter(self.search(environment=environment, last=True, limit=1)), None)

    def get_run_by_id(self, run_id: str) -> t.Optional[V1Run]:
        """
        Fetch a specific run by its identifier

        Args:
            run_id: The run identifier

        Returns:
            The run, or None if the identifier did not match any runs
        """
        try:
            return self.client.get_deserialized(model=V1Run, url_parts=[self.endpoint, run_id])
        except HttpClientError as e:
            if e.status_code == 404:
                return None
            raise e

    def search(
        self,
        environment: t.Optional[str] = None,
        last: bool = False,
        since_run_id: t.Optional[str] = None,
        sort: t.Literal["asc", "desc"] = "desc",
        limit: int = 100,
    ) -> t.List[V1Run]:
        """
        Search for runs based on the specified parameters

        Args:
            environment: Which environment to search in, or None for all environments
            last: Set to True to return just the most recent run that matches all the search terms
            since_run_id: Set to a run_id to return all the runs that happened since that run_id, or None to ignore
            sort: "asc" - oldest runs first, "desc" - newest runs first in the returned list
            limit: Maximum number of records to return

        Returns:
            A list of runs matching the search parameters
        """
        try:
            return self.client.get_many_deserialized(
                model=V1Run,
                url_parts=self.endpoint,
                params={
                    "environment": environment,
                    "last": last,
                    "since_run_id": since_run_id,
                    "sort": sort,
                    "limit": limit,
                },
            )
        except HttpClientError as e:
            if e.status_code == 404:
                return []
            raise e

    def get_evaluations_for_node(self, run_id: str, node_name: str) -> t.List[V1RunEvaluation]:
        """
        Fetch all the evaluations for a given DAG node within a given run.
        Note that there may be multiple evaluations for a single node if the scheduler broke up the work into batches
        or no evaluations at all if the scheduler skipped the node (which can happen if there are no intervals to run)

        Args:
            run_id: The run_id to search
            node_name: The name of the DAG node to fetch evaluations for

        Return:
            Any evaluations under the specified :run_id for the specifed :node_name
        """
        return self.get_evaluations(run_id, node_name)

    def get_evaluations(
        self, run_id: str, node_name: t.Optional[str] = None
    ) -> t.List[V1RunEvaluation]:
        """
        Fetch all the evaluations for a given DAG node within a given run.
        Note that there may be multiple evaluations for a single node if the scheduler broke up the work into batches
        or no evaluations at all if the scheduler skipped the node (which can happen if there are no intervals to run)

        Args:
            run_id: The run_id to search
            node_name: The name of the DAG node to fetch evaluations for, or None to fetch for all nodes

        Return:
            The list of matching evaluations
        """
        try:
            return self.client.get_many_deserialized(
                model=V1RunEvaluation,
                url_parts=[self.endpoint, run_id, "evaluations"],
                params={"node_name": node_name},
            )
        except HttpClientError as e:
            if e.status_code == 404:
                return []
            raise e


class V1EvaluationsApi(_Base):
    def get_evaluation_by_id(self, evaluation_id: str) -> t.Optional[V1EvaluationBase]:
        """
        Fetch an evaluation record by its ID.

        Args:
            evaluation_id: The id of the evaluation to fetch

        Return:
            The evaluation, or None if the supplied evaluation_id didnt match a record
        """
        try:
            return self.client.get_deserialized(
                model=V1EvaluationBase,
                url_parts=[self.endpoint, evaluation_id],
            )
        except HttpClientError as e:
            if e.status_code == 404:
                return None
            raise e

    def get_latest_evaluations_for_environment(self, environment: str) -> t.List[V1EvaluationBase]:
        """
        Snapshots can be evaluated multiple times (eg every time a cadence run triggers or when the snapshot backing
        an an incremental model is processed in batches).

        This returns the most recent evaluation record for each Snapshot in the specified environment.

        Args:
            environment: The environment to fetch evaluations for

        Return:
            A list of matching evaluation records. If the environment doesnt exist, an empty list is returned.
        """
        return self.search(environment=environment, latest=True, limit=None)

    def search(
        self,
        run_id: t.Optional[str] = None,
        plan_id: t.Optional[str] = None,
        environment: t.Optional[str] = None,
        latest: t.Optional[bool] = None,
        node_name: t.Optional[str] = None,
        sort: t.Literal["asc", "desc"] = "asc",
        limit: t.Optional[int] = 100,
    ) -> t.List[V1EvaluationBase]:
        try:
            return self.client.get_many_deserialized(
                model=V1EvaluationBase,
                url_parts=[self.endpoint],
                params={
                    "run_id": run_id,
                    "plan_id": plan_id,
                    "environment": environment,
                    "latest": latest,
                    "node_name": node_name,
                    "sort": sort,
                    "limit": limit,
                },
            )
        except HttpClientError as e:
            if e.status_code == 404:
                return []
            raise e


class V1ApiClient(_Base):
    def __init__(self, client: HttpClient):
        super().__init__(client=client, endpoint=urljoin(str(client._client.base_url), "/api/v1"))

    @cached_property
    def dags(self) -> V1DagApi:
        return V1DagApi(self.client, urljoin(self.endpoint, "dags"))

    @cached_property
    def runs(self) -> V1RunApi:
        return V1RunApi(self.client, urljoin(self.endpoint, "runs"))

    @cached_property
    def evaluations(self) -> V1EvaluationsApi:
        return V1EvaluationsApi(self.client, urljoin(self.endpoint, "evaluations"))



import json
import typing as t
from contextlib import contextmanager
from enum import Enum

import httpx
from httpx import Response
from httpx._types import PrimitiveData
from pydantic import BaseModel, ValidationError
from tenacity import retry, retry_if_exception, stop_after_attempt, wait_exponential

from tobikodata.helpers import ensure_list, urljoin


class HttpClientError(Exception):
    def __init__(self, *args: t.Any, status_code: t.Optional[int] = None, **kwargs: t.Any):
        super().__init__(*args, **kwargs)
        self.status_code = status_code


class HttpMethod(str, Enum):
    GET = "GET"
    PATCH = "PATCH"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"

    @property
    def is_get(self) -> bool:
        return self == HttpMethod.GET

    @property
    def is_post(self) -> bool:
        return self == HttpMethod.POST

    @property
    def is_put(self) -> bool:
        return self == HttpMethod.PUT

    @property
    def is_delete(self) -> bool:
        return self == HttpMethod.DELETE

    @property
    def is_patch(self) -> bool:
        return self == HttpMethod.PATCH


INPUT_MODEL = t.TypeVar("INPUT_MODEL", bound=BaseModel)


def retry_on_httpx_exception(exception: BaseException) -> bool:
    if isinstance(exception, httpx.TimeoutException):
        return True
    if (
        isinstance(exception, httpx.HTTPStatusError)
        and exception.response
        and exception.response.status_code >= 500
    ):
        return True
    if isinstance(exception, httpx.NetworkError):
        return True
    return False


class PublicHttpClient:
    """
    This HTTP client is intended for use against the public / stable API's (eg everything under /api/v1)
    It doesnt depend on any proprietary / server side code and is intended to be packaged along with the public API models
    separately from the rest of the codebase so it can be distributed independently without pulling in the rest of Enterprise
    """

    DEFAULT_TIMEOUT = 120

    def __init__(
        self,
        health_ready: t.Optional[str] = None,
        client: t.Optional[httpx.Client] = None,
        auth: t.Optional[httpx.Auth] = None,
        headers: t.Optional[t.Dict[str, str]] = None,
        error_class: t.Type[HttpClientError] = HttpClientError,
        retry_attempts: int = 5,
        wait_multiplier: float = 1.0,
        wait_min: int = 1,
        wait_max: int = 10,
    ):
        self._client = client or httpx.Client()
        if auth:
            self._client.auth = auth
        if headers:
            self._client.headers.update(headers)
        self.health_endpoint = health_ready
        self.error_class = error_class
        self.retry_attempts = retry_attempts
        self.wait_multiplier = wait_multiplier
        self.wait_min = wait_min
        self.wait_max = wait_max

    def _build_retry_decorator(self) -> t.Callable:
        return retry(
            retry=retry_if_exception(retry_on_httpx_exception),
            stop=stop_after_attempt(self.retry_attempts),
            wait=wait_exponential(
                multiplier=self.wait_multiplier, min=self.wait_min, max=self.wait_max
            ),
            reraise=True,
        )

    def _make_serializable(self, obj: t.Any) -> t.Any:
        # note: only required for objects that json.dumps() doesnt already know how to handle
        if isinstance(obj, BaseModel):
            return obj.model_dump(mode="json")
        elif isinstance(obj, Enum):
            return obj.value

        raise ValueError(f"Unsupported data type: {type(obj)}")

    def health_ready(self) -> None:
        if not self.health_endpoint:
            raise ValueError("No health endpoint configured")
        self.get(self.health_endpoint)

    def _call(
        self,
        method: HttpMethod,
        url: str,
        data: t.Optional[str] = None,
        params: t.Optional[t.Dict] = None,
        raise_for_status: bool = True,
        **kwargs: t.Any,
    ) -> Response:
        retry_decorator = self._build_retry_decorator()

        @retry_decorator
        def _call_retry(
            method: HttpMethod,
            url: str,
            data: t.Optional[str] = None,
            params: t.Optional[t.Dict] = None,
            raise_for_status: bool = True,
            **kwargs: t.Any,
        ) -> Response:
            if "timeout" not in kwargs:
                kwargs["timeout"] = self.DEFAULT_TIMEOUT
            if method.is_get:
                resp = self._client.get(url, params=params, **kwargs)
            elif method.is_post:
                resp = self._client.post(url, params=params, data=data, **kwargs)  # type: ignore
            elif method.is_put:
                resp = self._client.put(url, params=params, data=data, **kwargs)  # type: ignore
            elif method.is_delete:
                resp = self._client.delete(url, params=params, **kwargs)
            elif method.is_patch:
                resp = self._client.patch(url, params=params, data=data, **kwargs)  # type: ignore
            else:
                raise ValueError(f"Unsupported HTTP method: {method}")
            if raise_for_status:
                try:
                    resp.raise_for_status()
                except httpx.HTTPStatusError as e:
                    if "application/json" in resp.headers.get("Content-Type", ""):
                        detail = e.response.json().get("detail")
                        if detail:
                            raise self.error_class(
                                "\n"
                                + "\n".join(
                                    json.dumps(x, default=self._make_serializable)
                                    for x in ensure_list(detail)
                                ),
                                status_code=resp.status_code,
                            ) from e
                        raise e
                    else:
                        raise e
            return resp

        return _call_retry(method, url, data, params, raise_for_status, **kwargs)

    def _to_url(self, url_parts: t.Union[str, t.Iterable[str]]) -> str:
        return url_parts if isinstance(url_parts, str) else urljoin(*url_parts)

    def _to_data(
        self, data: t.Optional[t.Union[str, t.Sequence[BaseModel], t.Dict, BaseModel]]
    ) -> t.Optional[str]:
        if data is None:
            return None

        if isinstance(data, str):
            return data

        return json.dumps(data, default=self._make_serializable)

    def _to_params(
        self, params: t.Optional[t.Dict[str, t.Any]]
    ) -> t.Optional[t.Dict[str, PrimitiveData]]:
        if params is None:
            return None

        result = {}

        def _serialize_complex(obj: t.Any) -> t.Any:
            if isinstance(obj, t.get_args(PrimitiveData)):
                return obj
            return self._to_data(obj)

        for k, v in params.items():
            if v is None:
                continue

            # convert dict_values to list
            if isinstance(v, type({}.values())):
                v = list(v)

            # unwrap any collections
            if isinstance(v, (tuple, set, list)):
                v = [_serialize_complex(i) for i in v]
            else:
                v = _serialize_complex(v)

            result[k] = v

        return result

    @staticmethod
    def _deserialize_response(resp: Response, model: t.Type[INPUT_MODEL]) -> INPUT_MODEL:
        if resp.is_error:
            raise HttpClientError(f"Error response: {resp.text}")
        return model.model_validate(resp.json())

    @staticmethod
    def _deserialize_many_response(
        resp: Response, model: t.Type[INPUT_MODEL]
    ) -> t.List[INPUT_MODEL]:
        if resp.is_error:
            raise HttpClientError(f"Error response: {resp.text}")
        return [model.model_validate(obj) for obj in ensure_list(resp.json())]

    def patch(
        self,
        url_parts: t.Union[str, t.Iterable[str]],
        params: t.Optional[t.Dict] = None,
        data: t.Optional[t.Union[str, t.Sequence[BaseModel], t.Dict, BaseModel]] = None,
        raise_for_status: bool = True,
        **kwargs: t.Any,
    ) -> Response:
        return self._call(
            HttpMethod.PATCH,
            self._to_url(url_parts),
            params=self._to_params(params),
            data=self._to_data(data),
            raise_for_status=raise_for_status,
            **kwargs,
        )

    def post(
        self,
        url_parts: t.Union[str, t.Iterable[str]],
        params: t.Optional[t.Dict[str, t.Any]] = None,
        data: t.Optional[t.Union[str, t.Sequence[BaseModel], t.Dict, BaseModel]] = None,
        raise_for_status: bool = True,
        **kwargs: t.Any,
    ) -> Response:
        return self._call(
            HttpMethod.POST,
            self._to_url(url_parts),
            params=self._to_params(params),
            data=self._to_data(data),
            raise_for_status=raise_for_status,
            **kwargs,
        )

    @t.overload
    def post_deserialized(
        self,
        model: t.Type[INPUT_MODEL],
        url_parts: t.Union[str, t.Iterable[str]],
        params: t.Optional[t.Dict] = None,
        data: t.Optional[t.Union[str, t.Sequence[BaseModel], t.Dict, BaseModel]] = None,
        optional: t.Literal[False] = ...,
        raise_for_status: bool = True,
        **kwargs: t.Any,
    ) -> INPUT_MODEL: ...

    @t.overload
    def post_deserialized(
        self,
        model: t.Type[INPUT_MODEL],
        url_parts: t.Union[str, t.Iterable[str]],
        params: t.Optional[t.Dict] = None,
        data: t.Optional[t.Union[str, t.Sequence[BaseModel], t.Dict, BaseModel]] = None,
        optional: t.Literal[True] = ...,
        raise_for_status: bool = True,
        **kwargs: t.Any,
    ) -> t.Optional[INPUT_MODEL]: ...

    def post_deserialized(
        self,
        model: t.Type[INPUT_MODEL],
        url_parts: t.Union[str, t.Iterable[str]],
        params: t.Optional[t.Dict] = None,
        data: t.Optional[t.Union[str, t.Sequence[BaseModel], t.Dict, BaseModel]] = None,
        optional: bool = False,
        raise_for_status: bool = True,
        **kwargs: t.Any,
    ) -> t.Optional[INPUT_MODEL]:
        resp = self.post(
            url_parts,
            params=params,
            data=data,
            raise_for_status=raise_for_status,
            **kwargs,
        )
        try:
            return self._deserialize_response(resp, model)
        except ValidationError as e:
            if optional:
                return None
            raise HttpClientError(f"Expected {model.__name__}, got {resp.text}") from e

    def post_many_deserialized(
        self,
        model: t.Type[INPUT_MODEL],
        url_parts: t.Union[str, t.Iterable[str]],
        params: t.Optional[t.Dict] = None,
        data: t.Optional[t.Union[str, t.Sequence[BaseModel], t.Dict, BaseModel]] = None,
        raise_for_status: bool = True,
        **kwargs: t.Any,
    ) -> t.List[INPUT_MODEL]:
        resp = self.post(
            url_parts,
            params=params,
            data=data,
            raise_for_status=raise_for_status,
            **kwargs,
        )
        try:
            return self._deserialize_many_response(resp, model)
        except ValidationError as e:
            raise HttpClientError(f"Expected {model.__name__}, got {resp.text}") from e

    def get(
        self,
        url_parts: t.Union[str, t.Iterable[str]],
        params: t.Optional[t.Dict[str, t.Any]] = None,
        raise_for_status: bool = True,
        **kwargs: t.Any,
    ) -> Response:
        return self._call(
            HttpMethod.GET,
            self._to_url(url_parts),
            params=self._to_params(params),
            raise_for_status=raise_for_status,
            **kwargs,
        )

    @t.overload
    def get_deserialized(
        self,
        model: t.Type[INPUT_MODEL],
        url_parts: t.Union[str, t.Iterable[str]],
        params: t.Optional[t.Dict] = None,
        optional: t.Literal[False] = ...,
        raise_for_status: bool = True,
        **kwargs: t.Any,
    ) -> INPUT_MODEL: ...

    @t.overload
    def get_deserialized(
        self,
        model: t.Type[INPUT_MODEL],
        url_parts: t.Union[str, t.Iterable[str]],
        params: t.Optional[t.Dict] = None,
        optional: t.Literal[True] = ...,
        raise_for_status: bool = True,
        **kwargs: t.Any,
    ) -> t.Optional[INPUT_MODEL]: ...

    def get_deserialized(
        self,
        model: t.Type[INPUT_MODEL],
        url_parts: t.Union[str, t.Iterable[str]],
        params: t.Optional[t.Dict] = None,
        optional: bool = False,
        raise_for_status: bool = True,
        **kwargs: t.Any,
    ) -> t.Optional[INPUT_MODEL]:
        resp = self.get(
            url_parts,
            params=self._to_params(params),
            raise_for_status=raise_for_status,
            **kwargs,
        )
        try:
            return self._deserialize_response(resp, model)
        except ValidationError as e:
            if optional:
                return None
            raise HttpClientError(f"Expected {model.__name__}, got {resp.text}") from e

    def get_many_deserialized(
        self,
        model: t.Type[INPUT_MODEL],
        url_parts: t.Union[str, t.Iterable[str]],
        params: t.Optional[t.Dict] = None,
        raise_for_status: bool = True,
        **kwargs: t.Any,
    ) -> t.List[INPUT_MODEL]:
        resp = self.get(url_parts, params=params, raise_for_status=raise_for_status, **kwargs)
        try:
            return self._deserialize_many_response(resp, model)
        except ValidationError as e:
            raise HttpClientError(f"Expected {model.__name__}, got {resp.text}") from e

    def delete(
        self,
        url_parts: t.Union[str, t.Iterable[str]],
        params: t.Optional[t.Dict] = None,
        raise_for_status: bool = True,
        **kwargs: t.Any,
    ) -> Response:
        return self._call(
            HttpMethod.DELETE,
            self._to_url(url_parts),
            params=self._to_params(params),
            raise_for_status=raise_for_status,
            **kwargs,
        )

    def delete_deserialized(
        self,
        model: t.Type[INPUT_MODEL],
        url_parts: t.Union[str, t.Iterable[str]],
        params: t.Optional[t.Dict] = None,
        raise_for_status: bool = True,
        **kwargs: t.Any,
    ) -> INPUT_MODEL:
        resp = self.delete(url_parts, params=params, raise_for_status=raise_for_status, **kwargs)
        try:
            return self._deserialize_response(resp, model)
        except ValidationError as e:
            raise HttpClientError(f"Expected {model.__name__}, got {resp.text}") from e

    def delete_many_deserialized(
        self,
        model: t.Type[INPUT_MODEL],
        url_parts: t.Union[str, t.Iterable[str]],
        params: t.Optional[t.Dict] = None,
        raise_for_status: bool = True,
        **kwargs: t.Any,
    ) -> t.List[INPUT_MODEL]:
        resp = self.delete(url_parts, params=params, raise_for_status=raise_for_status, **kwargs)
        try:
            return self._deserialize_many_response(resp, model)
        except ValidationError as e:
            raise HttpClientError(f"Expected {model.__name__}, got {resp.text}") from e

    def put(
        self,
        url_parts: t.Union[str, t.Iterable[str]],
        data: t.Optional[t.Union[str, t.Sequence[BaseModel], t.Dict, BaseModel]] = None,
        params: t.Optional[t.Dict] = None,
        raise_for_status: bool = True,
        **kwargs: t.Any,
    ) -> Response:
        return self._call(
            HttpMethod.PUT,
            self._to_url(url_parts),
            data=self._to_data(data),
            params=self._to_params(params),
            raise_for_status=raise_for_status,
            **kwargs,
        )

    def put_deserialized(
        self,
        model: t.Type[INPUT_MODEL],
        url_parts: t.Union[str, t.Iterable[str]],
        data: t.Optional[t.Union[str, t.Sequence[BaseModel], t.Dict, BaseModel]] = None,
        params: t.Optional[t.Dict] = None,
        raise_for_status: bool = True,
        **kwargs: t.Any,
    ) -> INPUT_MODEL:
        resp = self.put(
            url_parts,
            data=data,
            params=params,
            raise_for_status=raise_for_status,
            **kwargs,
        )
        try:
            return self._deserialize_response(resp, model)
        except ValidationError as e:
            raise HttpClientError(f"Expected {model.__name__}, got {resp.text}") from e

    @contextmanager
    def stream(
        self,
        method: HttpMethod,
        url_parts: t.Union[str, t.Iterable[str]],
        data: t.Optional[t.Union[str, t.Sequence[BaseModel], t.Dict, BaseModel]] = None,
        params: t.Optional[t.Dict] = None,
        raise_for_status: bool = True,
        **kwargs: t.Any,
    ) -> t.Iterator[Response]:
        with self._client.stream(
            method=method.value,
            url=self._to_url(url_parts),
            data=self._to_data(data),  # type: ignore
            params=self._to_params(params),
            **kwargs,
        ) as r:
            yield r

    def close(self) -> None:
        self._client.close()


class BearerAuth(httpx.Auth):
    def __init__(self, token: str) -> None:
        self.token = token

    def auth_flow(self, request: httpx.Request) -> t.Generator[httpx.Request, httpx.Response, None]:
        request.headers["Authorization"] = f"Bearer {self.token}"
        yield request



import typing as t
from datetime import datetime

from pydantic import BaseModel, HttpUrl

from tobikodata.http_client.api_models.v1.common import V1Status


class V1Run(BaseModel):
    environment: str
    run_id: str
    start_at: t.Optional[datetime]
    end_at: t.Optional[datetime]
    error_message: t.Optional[str]
    status: V1Status
    link: HttpUrl

    @property
    def complete(self) -> bool:
        return self.status.complete



import typing as t
from datetime import datetime
from functools import cached_property

from pydantic import BaseModel, HttpUrl

from tobikodata.http_client.api_models.v1.common import V1Status


class V1AuditEvaluation(BaseModel):
    name: str
    execution_time: datetime
    interval_start: datetime
    interval_end: datetime
    status: t.Literal[V1Status.SUCCESS, V1Status.FAILED]
    link: HttpUrl
    log_link: HttpUrl
    log: t.Optional[str]


class V1EvaluationBase(BaseModel):
    evaluation_id: str
    node_name: str
    start_at: t.Optional[datetime]
    end_at: t.Optional[datetime]
    error_message: t.Optional[str]
    log: t.Optional[str]
    status: V1Status
    link: HttpUrl
    log_link: HttpUrl
    audits: t.List[V1AuditEvaluation] = []

    @property
    def complete(self) -> bool:
        return self.status.complete

    @cached_property
    def audits_by_name(self) -> t.Mapping[str, V1AuditEvaluation]:
        return {a.name: a for a in self.audits}


class V1PlanEvaluation(V1EvaluationBase):
    plan_id: str


class V1RunEvaluation(V1EvaluationBase):
    run_id: str


V1Evaluation = t.Union[V1PlanEvaluation, V1RunEvaluation]



from tobikodata.http_client.public import (
    BearerAuth as BearerAuth,
)
from tobikodata.http_client.public import (
    HttpClientError as HttpClientError,
)
from tobikodata.http_client.public import (
    PublicHttpClient as HttpClient,
)

# this needs to appear last to prevent a circular import
from tobikodata.http_client.api import V1ApiClient as V1ApiClient  # isort: skip

__all__ = ["BearerAuth", "HttpClientError", "HttpClient", "V1ApiClient"]



import typing as t
from datetime import datetime
from enum import Enum

from pydantic import BaseModel, ConfigDict, HttpUrl


class V1DagNodeType(str, Enum):
    MODEL = "model"
    AUDIT = "audit"


class V1DagNodeSourceType(str, Enum):
    SQL = "sql"
    SEED = "seed"
    PYTHON = "python"
    EXTERNAL = "external"
    AUDIT = "audit"


class V1DagNode(BaseModel):
    # Prevents warning about property starting with model_
    model_config = ConfigDict(
        protected_namespaces=(),
    )

    name: str
    model_name: str
    description: t.Optional[str]
    source_type: V1DagNodeSourceType
    tags: t.List[str]
    parent_names: t.List[str]
    audit_names: t.List[str]
    type: V1DagNodeType
    link: HttpUrl


class V1Dag(BaseModel):
    environment: str
    start_at: datetime
    schedule_seconds: int
    schedule_cron: str
    nodes: t.List[V1DagNode]
    link: HttpUrl




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: tobikodata/http_client:dist

from setuptools import setup

setup(**{
    'author': 'TobikoData Inc.',
    'author_email': 'engineering@tobikodata.com',
    'install_requires': (
        'Authlib',
        'httpx',
        'pydantic>=2.0.0',
        'rich[jupyter]',
        'ruamel.yaml',
        'tenacity',
        'tobiko-cloud-helpers==202506.39.0',
    ),
    'name': 'tobiko-cloud-api-client',
    'namespace_packages': (
    ),
    'package_data': {
    },
    'packages': (
        'tobikodata',
        'tobikodata.http_client',
        'tobikodata.http_client.api_models.v1',
    ),
    'python_requires': '<3.13,>=3.9',
    'version': '202506.39.0',
})
